import random
import time


from Numberjack import *
from Alphabet import *
from EnigmaEmulator import *




def enumerate_reversed(L):
    for index in reversed(xrange(len(L))):
          yield index, L[index]

 
def singleRotorEncode(alphabet, rotor, length): # data: single rotor, a plaintext. Emulate the encoding
    N = alphabet.size()
    P = VarArray([Variable(r,r) for r in rotor])
    
    plaintext = VarArray(length, 0, N-1, 'text') 
    ciphertext = VarArray(length, 0, N-1,'cipher')
    key = Variable(0, N-1, 'key')
    
    model = Model(
        [c == P[(key + t + i) % N] for i,(c,t) in enumerate(zip(plaintext,ciphertext))]
    ) 
    
    solver = model.load("Mistral2")
    solver.setVerbosity(0)
    solver.startNewSearch()
    for i in range(1000):
        if solver.getNextSolution():
            print ''.join([alphabet.char[t.get_value()] for t in plaintext]), \
             '-[%i]->'%(key.get_value()), \
             ''.join([alphabet.char[c.get_value()] for c in ciphertext])
        else:
            break
            
            
def singleRotorBreak(alphabet, rotor, code, crib): # data: single rotor, a plaintext. Break the encoding
    N = alphabet.size()
    R = VarArray([Variable(r,r,'r_%i'%i) for i,r in enumerate(rotor)])

    plaintext = VarArray(len(code), 0, N-1, 'text')
    ciphertext = VarArray(len(code), 0, N-1, 'cipher')
    key = Variable(0, N-1, 'key')

    model = Model(
        [xc == R[(key + xt + i) % N] for i,(xt,xc) in enumerate(zip(plaintext,ciphertext))],
        [xc == alphabet.digit[c] for xc,c in zip(ciphertext,code)],
        [xt == alphabet.digit[t] for xt,t in zip(plaintext,crib)]
    )

    solver = model.load("Mistral2")
    solver.setVerbosity(0)
    ok = solver.solve()
    while ok:
        print ''.join([alphabet.char[t.get_value()] for t in plaintext]), \
        '-[%i]->'%(key.get_value()), \
        ''.join([alphabet.char[c.get_value()] for c in ciphertext])
        ok = solver.getNextSolution()
    
    
def multipleRotorsBreak(alphabet, rotors, code, crib): # data: a set of sequential rotors, a plaintext. Break the encoding
    N = alphabet.size()
    
    Rotors = [VarArray([Variable(r,r,'r_%i'%i) for i,r in enumerate(R)]) for R in rotors]

    plaintext = VarArray(len(code), 0, N-1, 'text')
    ciphertext = [VarArray(len(code), 0, N-1, 'cipher_%i'%i) for i in range(len(rotors))]
    key = VarArray(len(rotors), 0, N-1, 'key')

    model = Model(
        [xc == alphabet.digit[c] for xc,c in zip(ciphertext[-1],code)],
        [xt == alphabet.digit[t] for xt,t in zip(plaintext,crib)]
    )
    
    previous = plaintext
    for j,(k,R,cipher) in enumerate(zip(key,Rotors,ciphertext)):
        model.add( [xc == R[(k + xt + (i//(N**j))) % N] for i,(xt,xc) in enumerate(zip(previous,cipher))] )
        previous = cipher
    

    solver = model.load("Mistral2")
    solver.setVerbosity(0)
    ok = solver.solve()
    while ok:
        print ''.join([alphabet.char[t.get_value()] for t in plaintext]), \
        '-[%s]->'%(' '.join([str(k.get_value()) for k in key])), \
        ''.join([alphabet.char[c.get_value()] for c in ciphertext[-1]])
        ok = solver.getNextSolution()
    

def commercialEnigmaEncode(alphabet, inputtext, enigma, key): # data: a commercial enigma machine, a plaintext. Emulate the encoding
    N = alphabet.size()
    
    ForwardRotors = [VarArray([Variable(r,r,'fr_%i'%i) for i,r in enumerate(R.forward)]) for R in enigma.rotors]
    BackwardRotors = [VarArray([Variable(r,r,'br_%i'%i) for i,r in enumerate(R.backward)]) for R in enigma.rotors]
    
    Reflector = VarArray([Variable(r,r,'r_%i'%i) for i,r in enumerate(enigma.reflector)])

    plaintext = VarArray(len(inputtext), 0, N-1, 'text')
    reflected = VarArray(len(inputtext), 0, N-1, 'reflected')
    forward_cipher = [VarArray(len(inputtext), 0, N-1, 'forward_%i'%i) for i in range(len(enigma.rotors))]
    backward_cipher = [VarArray(len(inputtext), 0, N-1, 'backward_%i'%i) for i in range(len(enigma.rotors))]
    key_var = VarArray(len(enigma.rotors), 0, N-1, 'key')

    model = Model(
        [xt == alphabet.digit[t] for xt,t in zip(plaintext,inputtext)],
        [xk == k for xk,k in zip(key_var,key)]
    )
    
    previous = plaintext
    for j,(k,R,cipher) in enumerate(zip(key_var,ForwardRotors,forward_cipher)):
        model.add( [x_next == R[(x_prev + k + (i//(N**j))) % N] for i,(x_prev,x_next) in enumerate(zip(previous,cipher))] )
        previous = cipher
       
    model.add( [x_next == Reflector[x_prev] for i,(x_prev,x_next) in enumerate(zip(forward_cipher[-1],reflected))] )

    previous = reflected
    for j,(k,R,cipher) in enumerate_reversed(zip(key_var,ForwardRotors,backward_cipher)):
        model.add( [x_prev == R[(k + x_next + (i//(N**j))) % N] for i,(x_prev,x_next) in enumerate(zip(previous,cipher))] )
        previous = cipher
    

    solver = model.load("Mistral2")
    solver.setVerbosity(0)
    if solver.solve():
        return ''.join([alphabet.char[c.get_value()] for c in backward_cipher[0]])
        
        
def commercialEnigmaDecode(alphabet, code, enigma, key): # data: a commercial enigma machine, a plaintext. Emulate the encoding
    N = alphabet.size()
    
    ForwardRotors = [VarArray([Variable(r,r,'fr_%i'%i) for i,r in enumerate(R.forward)]) for R in enigma.rotors]
    BackwardRotors = [VarArray([Variable(r,r,'br_%i'%i) for i,r in enumerate(R.backward)]) for R in enigma.rotors]
    
    Reflector = VarArray([Variable(r,r,'r_%i'%i) for i,r in enumerate(enigma.reflector)])

    plaintext = VarArray(len(code), 0, N-1, 'text')
    reflected = VarArray(len(code), 0, N-1, 'reflected')
    forward_cipher = [VarArray(len(code), 0, N-1, 'forward_%i'%i) for i in range(len(enigma.rotors))]
    backward_cipher = [VarArray(len(code), 0, N-1, 'backward_%i'%i) for i in range(len(enigma.rotors))]
    key_var = VarArray(len(enigma.rotors), 0, N-1, 'key')

    model = Model(
        [xt == alphabet.digit[t] for xt,t in zip(backward_cipher[0],code)],
        [xk == k for xk,k in zip(key_var,key)]
    )
    
    previous = plaintext
    for j,(k,R,cipher) in enumerate(zip(key_var,ForwardRotors,forward_cipher)):
        model.add( [x_next == R[(x_prev + k + (i//(N**j))) % N] for i,(x_prev,x_next) in enumerate(zip(previous,cipher))] )
        previous = cipher
       
    model.add( [x_next == Reflector[x_prev] for i,(x_prev,x_next) in enumerate(zip(forward_cipher[-1],reflected))] )

    previous = reflected
    for j,(k,R,cipher) in enumerate_reversed(zip(key_var,ForwardRotors,backward_cipher)):
        model.add( [x_prev == R[(k + x_next + (i//(N**j))) % N] for i,(x_prev,x_next) in enumerate(zip(previous,cipher))] )
        previous = cipher
    

    solver = model.load("Mistral2")
    solver.setVerbosity(0)
    if solver.solve():
        return ''.join([alphabet.char[c.get_value()] for c in plaintext])
        
        
def commercialEnigmaBreak(alphabet, code, enigma, crib): # data: a commercial enigma machine, a plaintext. Emulate the encoding
    N = alphabet.size()
    
    ForwardRotors = [VarArray([Variable(r,r,'fr_%i'%i) for i,r in enumerate(R.forward)]) for R in enigma.rotors]
    BackwardRotors = [VarArray([Variable(r,r,'br_%i'%i) for i,r in enumerate(R.backward)]) for R in enigma.rotors]
    
    Reflector = VarArray([Variable(r,r,'r_%i'%i) for i,r in enumerate(enigma.reflector)])

    plaintext = VarArray(len(crib), 0, N-1, 'text')
    reflected = VarArray(len(crib), 0, N-1, 'reflected')
    forward_cipher = [VarArray(len(crib), 0, N-1, 'forward_%i'%i) for i in range(len(enigma.rotors))]
    backward_cipher = [VarArray(len(crib), 0, N-1, 'backward_%i'%i) for i in range(len(enigma.rotors))]
    key_var = VarArray(len(enigma.rotors), 0, N-1, 'key')

    model = Model(
        [xt == alphabet.digit[t] for xt,t in zip(backward_cipher[0],code)],
        [xt == alphabet.digit[t] for xt,t in zip(plaintext,crib)]
    )
    
    previous = plaintext
    for j,(k,R,cipher) in enumerate(zip(key_var,ForwardRotors,forward_cipher)):
        model.add( [x_next == R[(x_prev + k + (i//(N**j))) % N] for i,(x_prev,x_next) in enumerate(zip(previous,cipher))] )
        previous = cipher
       
    model.add( [x_next == Reflector[x_prev] for i,(x_prev,x_next) in enumerate(zip(forward_cipher[-1],reflected))] )

    previous = reflected
    for j,(k,R,cipher) in enumerate_reversed(zip(key_var,ForwardRotors,backward_cipher)):
        model.add( [x_prev == R[(k + x_next + (i//(N**j))) % N] for i,(x_prev,x_next) in enumerate(zip(previous,cipher))] )
        previous = cipher
    

    solver = model.load("Mistral2")
    solver.setVerbosity(0)
    
    possible_keys = []
    start = time.clock()
    ok = solver.solve()
    while ok:
        solution = [k.get_value() for k in key_var]
        possible_keys.append(solution)
        print 'key = %8s, %8inds, %.*fs'%('/'.join([str(k) for k in solution]),solver.getNodes(),2,(time.clock()-start))
        # print ''.join([alphabet[c.get_value()] for c in text]), '%8inds,'%solver.getNodes(), '%.*fs'%(2,(time.clock()-start))
        ok = solver.getNextSolution()
    print 'total = %8inds, %.*fs'%(solver.getNodes(),2,(time.clock()-start))
    return possible_keys

           
            
if __name__ == '__main__':
    
    random.seed(12345)
    
    alphabet = Alphabet()
    
    perm = ([15, 14, 5, 22, 24, 13, 25, 12, 18, 9, 1, 23, 20, 16, 17, 21, 7, 2, 3, 11, 4, 8, 6, 19, 0, 10], \
            [21, 17, 13, 6, 0, 16, 10, 20, 3, 9, 14, 19, 4, 7, 2, 8, 12, 5, 22, 23, 24, 18, 25, 11, 15, 1], \
            [8, 2, 16, 25, 15, 7, 1, 4, 22, 9, 17, 11, 24, 12, 5, 18, 20, 13, 6, 19, 0, 3, 10, 23, 21, 14])
            
    reflector = [10, 18, 16, 6, 15, 19, 3, 12, 20, 23, 0, 14, 7, 22, 11, 4, 2, 24, 1, 5, 8, 25, 13, 9, 17, 21]
    
    
    print 'EXERCICE 1\n'
    singleRotorEncode(alphabet, perm[0], 3)



    print '\n\nEXERCICE 2\n'
    cipher0 = 'LCKPYBKOUMPVHKZNBITLIKUXOVN'

    """
    The ciphertext above was gereated as follows:

    k = ?
    text = 'H??????????????????????????'
    print ''.join([alphabet[perm1[(digit[t]+i+k) % N]] for i,t in enumerate(text)])
    """

    singleRotorBreak(alphabet, perm[0], cipher0, 'H')


    print '\n\nEXERCICE 3\n'
    cipher1 = 'YTHXGIPIBUZYCCQSLPELWDRAOZXOSGOHCZNOLRGYXFYZTWFMZIBDBUVOU'

    rotors = [Rotor(perm=perm[0]), Rotor(perm=perm[1])]


    """
    The ciphertext above was gereated as follows:

    text = 'TH???????????????????????????????????????????????????????'
    rotors[0].set_key(12)
    rotors[1].set_key(7)

    cipher = ''
    for i,c in enumerate(text):
        c = digit[c]
        for rotor in rotors:
            c = rotor.encrypt_forward(c)
        j = 0
        while j<len(rotors) and rotors[j].step():
            j += 1
        cipher += alphabet[c]

    print cipher
    """

    multipleRotorsBreak(alphabet, [perm[0]], cipher0, 'H')
    print
    multipleRotorsBreak(alphabet, [r.forward for r in rotors], cipher1, 'TH')


    print '\n\nEXERCICE 4'
    enigma = Machine(alphabet, rotors=[Rotor(p) for p in perm], plug_board=range(alphabet.size()), reflector=reflector)

    T = 'TESTFORMYENIGMACONSTRAINTPROGRAMMINGMODEL'
    N = alphabet.size()

    for i in range(1):
        key = [random.randint(0,N-1) for j in range(len(enigma.rotors))]
        key = [3,4,5]

        print

        for l in range(1,8):
            enigma.set_key(key)
            ciphertext = enigma.lencrypt(T,l)
            print 'Enigma  cipher:', ciphertext
        # ciphertext = enigma.encrypt(T)
        # print 'Enigma  cipher:', ciphertext
        C = commercialEnigmaEncode(alphabet, T, enigma, key)
        print 'CPmodel cipher:', C
        assert(C == ciphertext)

        print

        enigma.set_key(key)
        plaintext = enigma.encrypt(ciphertext)
        print 'Enigma  plaintext:', plaintext
        T = commercialEnigmaEncode(alphabet, ciphertext, enigma, key)
        print 'CPmodel plaintext:', T
        assert(T == plaintext)

    
    
    print '\n\nEXERCICE 5\n'
    cipher2 = 'SLIAMQPKAYMDNZPPJUWWWTNFCJZEHRNQOEANJGVBMHHDZBXMACJRZYPDXJUSWTGOPCJQUJCZVAMEDUPAIMPCXP'
 
    """
    The code above was gereated as follows:

    enigma = Machine(alphabet, rotors=[Rotor(p) for p in perm], plug_board=range(N), reflector=reflector)
        
    enigma.set_key([?,?,?])
    cipher2 = enigma.encrypt('WELL??????????????????????????????????????????????????????????????????????????????????')

    print cipher2
    """
    
    enigma = Machine(alphabet, rotors=[Rotor(p) for p in perm], plug_board=range(N), reflector=reflector)
    
    possible_keys = commercialEnigmaBreak(alphabet, cipher2, enigma, 'WEL')
    
    print 
    
    for keys in possible_keys:
        enigma.set_key(keys)
        print 'try key %8s: %s'%('/'.join([str(k) for k in keys]), enigma.encrypt(cipher2))
    
    
    
    
    