#! /usr/bin/env python

import random
import pickle

from Alphabet import *


class Rotor(object):
    
    def __init__(self, perm, key=None):
        N = len(perm)
        self.set_key(random.randint(0,N-1) if key is None else key)
        self.forward = perm[::]
        self.backward = [i for c,i in sorted(zip(self.forward, range(N)))]
        
    def size(self):
        return len(self.forward)
        
    def set_key(self, k):
        self.position = k
        self.init_position = k
    
    def step(self):
        self.position = (self.position + 1) % self.size()
        return self.position == self.init_position
        
    def encrypt_forward(self, c):
        return self.forward[(self.position + c) % self.size()]
        
    def encrypt_backward(self, c):
        return (self.backward[c] - self.position) % self.size()
        

    
def random_symmetric_permutation(k,N): # symmetric permutation of 2*k elements out of N (N-2*k elements are not permuted)
    assert 2*k <= N

    m = range(N)
    random.shuffle(m)
    p = [-1]*N
    for a,b in zip(m[:k], m[k:]):
        p[a] = b
        p[b] = a
    return [p[i] if p[i] >= 0 else i for i in range(N)]
        
        
class Machine(object):
    
    def __init__(self, alphabet, rotors, plug_board, reflector):
        self.alphabet = alphabet
        self.rotors = rotors
        self.reflector = reflector[::] 
        self.plug_board = plug_board[::]
            
    def get_key(self):
        return [r.position for r in self.rotors]
        
    def set_key(self, keys):
        for r,k in zip(self.rotors, keys):
            r.position = k
            r.init_position = k
            
    def set_plugs(self, plugs):
        self.plug_board = [p for p in plugs]
        
    def step(self):
        i = 0
        while i < len(self.rotors) and self.rotors[i].step():
            i += 1
    
    def encryption(self, c):
        c = self.plug_board[c]

        for r in self.rotors:
            c = r.encrypt_forward(c)
            
        c = self.reflector[c]
        
        for r in self.rotors[::-1]:
            c = r.encrypt_backward(c)
            
        return self.plug_board[c]
                
    def encrypt_char(self, c):
        c = self.encryption(self.alphabet.digit[c])   
        self.step()
        return self.alphabet.char[c]
        
    def encrypt(self, plaintext):
        return ''.join([self.encrypt_char(c) if c in self.alphabet.known else c for c in plaintext])

        
    def limited_encryption(self, c, limit):
        c = self.plug_board[c]

        count = 0

        for r in self.rotors:
            c = r.encrypt_forward(c)
            count += 1
            if count == limit:
                return c

        c = self.reflector[c]
        count += 1
        if count == limit:
            return c

        for r in self.rotors[::-1]:
            c = r.encrypt_backward(c)
            count += 1
            if count == limit:
                return c

        return self.plug_board[c]

    def lencrypt_char(self, c, limit):
        c = self.limited_encryption(self.alphabet.digit[c], limit)
        self.step()
        return self.alphabet.char[c]

    def lencrypt(self, plaintext, limit):
        return ''.join([self.lencrypt_char(c, limit) if c in self.alphabet.known else c for c in plaintext])
        

        
if __name__ == '__main__':
    
    import argparse
    
    parser = argparse.ArgumentParser(description='Enigma Emulator')
    parser.add_argument('file',type=str,help='path to text file')
    parser.add_argument('--seed',type=int,default=12345,help='random seed')
    parser.add_argument('--machine',type=str,default='machine.txt',help='machine file')
    parser.add_argument('--key',type=str,default=None,help='key file')
    parser.add_argument('--nrotors',type=int,default=3,help='number of rotors')
    parser.add_argument('--nplugs',type=int,default=10,help='number of switches')


    args = parser.parse_args()

    text = open(args.file, 'r').read()
    
    A = Alphabet()
    
    engima = None
    try:
        enigma = pickle.load(open(args.machine, 'r'))
        print 'machine loaded from %s'%args.machine
    except:
        print 'creating a new machine: %s'%args.machine
        perms = [range(A.size()) for i in range(args.nrotors)]
        for p in perms:
            random.shuffle(p)
        reflector = random_symmetric_permutation(A.size()/2, A.size())
        plug_board = random_symmetric_permutation(min(A.size()/2,10), A.size())
        enigma = Machine(A, [Rotor(p) for p in perms], plug_board, reflector)
        pickle.dump(enigma, open(args.machine, 'w'))
        try:
            keyfile = open(args.key, 'w')
            print 'creating a new key file: %s'%args.key
            pickle.dump(enigma.get_key(), keyfile)
            pickle.dump(enigma.plug_board, keyfile)
        except:
            pass
        
    if  args.key is not None:
        try: 
            keyfile = open(args.key, 'r')
            enigma.set_key(pickle.load(keyfile))
            enigma.set_plugs(pickle.load(keyfile))
            print 'key loaded from %s'%args.key
        except:
            pass
    

    
    print enigma.encrypt(text)
        
    
    
        
        
        
    
    

    
    
    
   
